/********** tell emacs we use -*- c++ -*- style comments *******************
 * $Revision: 1.1 $  $Author: reids $  $Date: 2001/03/16 17:56:01 $
 *
 * PROJECT:      Distributed Robotic Agents
 * DESCRIPTION:  
 *
 * (c) Copyright 2001 CMU. All rights reserved.
 ***************************************************************************/

#ifndef INCParsePos_h
#define INCParsePos_h

#include "RefCounter.h"
#include <string>
#include <vector>

struct ParseFilePos {
  string fileName;
  int lineNumber;

  ParseFilePos(string _fileName, int _lineNumber)
    : fileName(_fileName), lineNumber(_lineNumber) { }
};

struct ParsePosSharedInfo : public RefCounter {
  vector<ParseFilePos> stack;
  string fileName;

  ParsePosSharedInfo(string _fileName)
    : fileName(_fileName) { }
};

struct ParsePos {
  string beforeToken;
  int lineNumber;
  ParsePosSharedInfo *sharedInfo;

  ParsePos(void) { sharedInfo = 0; }
  ParsePos(string _beforeToken, int _lineNumber,
	   ParsePosSharedInfo *_sharedInfo)
    : beforeToken(_beforeToken),
      lineNumber(_lineNumber),
      sharedInfo(_sharedInfo) {
    sharedInfo->ref();
  }
  ParsePos(const ParsePos &pr)
    : beforeToken(pr.beforeToken),
      lineNumber(pr.lineNumber),
      sharedInfo(pr.sharedInfo) {
    sharedInfo->ref();
  }
  ~ParsePos(void) {
    if (0 != sharedInfo) sharedInfo->unref();
  }

  static void putLineDirective(ostream &out, int _lineNumber,
			       const char *_fileName,
			       bool enteringFile = true);
  void putLineDirective(ostream &out, bool enteringFile = true);
};

#endif // INCParsePos_h

/***************************************************************************
 * REVISION HISTORY:
 * $Log: ParsePos.h,v $
 * Revision 1.1  2001/03/16 17:56:01  reids
 * Release of Trey's code to generate IPC format strings from XDR definitions.
 *
 * Revision 1.2  2001/02/06 02:01:54  trey
 * fixed bugs discovered while generating commonTypes message formats
 *
 * Revision 1.1  2001/02/05 21:10:46  trey
 * initial check-in
 *
 *
 ***************************************************************************/
